import { pickCert } from './check-cert';
import { Parser } from './parser';
import { getSiteNames } from './utility';
import md5 from 'nano-md5';
import fs from 'fs';

export interface SiteHttps {
  ports: number[];
  cert: string;
  redirect?: boolean;
  hsts?: boolean;
}

export interface Header {
  name: string;
  value: string;
}

export interface SiteRenderData {
  domains: string[];
  ports: number[];
  https?: SiteHttps;
  headers?: Header[];
  normalizeDomain?: string;
  disableTop?: boolean;
  serverExtra?: string[];
  locationExtra?: string[];
  htpasswd?: string;
}

export interface ProxyRenderData extends SiteRenderData {
  proxy: true;
  minio?: boolean;
  timeout?: number;
  upstream: string;
  noVerifyCerts?: boolean;
  noBuffer?: boolean;
  sni?: boolean;
  sniName?: string;
  hostHeader?: string;
  noCache?: boolean;
}

interface FileRenderData extends SiteRenderData {
  root: string;
  index: string;
  expires: string;
}

export interface StaticRenderData extends FileRenderData {
  static: true;
  browse?: boolean;
}

export interface PhpRenderData extends FileRenderData {
  php: true;
  upstream: string;
  phpExtra?: string[];
}

export interface RedirectRenderData extends SiteRenderData {
  redirect: true;
  upstream: string;
  code: number;
}

export type SpecificRenderData =
  | PhpRenderData
  | StaticRenderData
  | ProxyRenderData
  | RedirectRenderData;

export interface RenderData {
  purgeAllowed?: string[];
  externalRealIp?: boolean;
  realIp?: string[];
  limitRate?: string;
  limitBurst?: string;
  maxCacheSize: string;
  dhparamPath: string;
  ticketKeyPath: string;
  certsPath: string;
  sites: SiteRenderData[];
  httpExtra?: string[];
  nginxExtra?: string[];
  cors?: boolean;
}

async function getSiteData(
  domain: string,
  input: Record<string, string> = process.env,
): Promise<SpecificRenderData> {
  const parser = new Parser(`SITE_${domain}_`, input);
  let https: SiteHttps;
  const httpsCert = parser.getString('HTTPS');
  const domains = domain.split('+');
  if (httpsCert !== '0' && httpsCert !== 'false') {
    const cert =
      !httpsCert ||
      httpsCert === '1' ||
      httpsCert === 'true' ||
      httpsCert === 'auto'
        ? pickCert(domains)
        : httpsCert;
    if (cert) {
      https = {
        cert,
        ports: parser.getArrayNumber('HTTPS_PORT') || [443],
        redirect: !parser.getBoolean('HTTPS_NOREDIR'),
        hsts: parser.getBoolean('HSTS'),
      };
    }
  }

  const targetUrl = new URL(input[`SITE_${domain}`]);
  let specificRenderData: SpecificRenderData;
  if (targetUrl.protocol === 'static:') {
    specificRenderData = {
      static: true,
      root: targetUrl.pathname,
      index: parser.getString('INDEX') || 'index.html',
      expires: parser.getString('EXPIRES') || '10m',
      browse: parser.getBoolean('BROWSE'),
    } as StaticRenderData;
  } else if (targetUrl.protocol === 'php:') {
    specificRenderData = {
      php: true,
      root: targetUrl.pathname,
      index: parser.getString('INDEX') || 'index.php',
      expires: parser.getString('EXPIRES') || '10m',
      upstream: targetUrl.host,
      phpExtra: parser.getArray('PHP_EXTRA'),
    } as PhpRenderData;
  } else if (targetUrl.protocol.startsWith('redirect+')) {
    specificRenderData = {
      redirect: true,
      upstream: targetUrl.href.slice(9),
      code: parser.getNumber('REDIRECT_CODE') || 301,
    } as RedirectRenderData;
  } else {
    const sni = parser.getString('SNI');
    specificRenderData = {
      proxy: true,
      upstream: input[`SITE_${domain}`],
      noVerifyCerts: parser.getBoolean('NO_VERIFY_CERTS'),
      noBuffer: parser.getBoolean('NO_BUFFER'),
      sni: sni === '1',
      sniName: sni === '1' ? undefined : sni,
      hostHeader: parser.getString('HOST'),
      noCache: parser.getBoolean('NO_CACHE'),
      minio: parser.getBoolean('MINIO'),
      timeout: parser.getNumber('TIMEOUT'),
    } as ProxyRenderData;
  }

  const basicPasswords = Object.entries(parser.getDict('HTPASSWD'));
  if (basicPasswords.length) {
    const htpasswd = basicPasswords
      .map(([user, pass]) => `${user}:${md5.crypt(pass)}`)
      .join('\n');
    await fs.promises.writeFile(
      `/etc/nginx/generated/htpasswd-${domain}`,
      htpasswd,
    );
  }

  return {
    domains: domains,
    ports: parser.getArrayNumber('PORT') || [80],
    https,
    headers: Object.entries(parser.getDict('HEADER')).map(([name, value]) => ({
      name: name.replace(/_/g, '-'),
      value,
    })),
    normalizeDomain: parser.getString('NORMALIZE_DOMAIN'),
    disableTop: parser.getBoolean('DISABLE_TOP'),
    serverExtra: parser.getArray('SERVER_EXTRA'),
    locationExtra: parser.getArray('LOCATION_EXTRA'),
    htpasswd: basicPasswords.length
      ? `/etc/nginx/generated/htpasswd-${domain}`
      : undefined,
    ...specificRenderData,
  };
}

export async function getData(
  input: Record<string, string> = process.env,
): Promise<RenderData> {
  const parser = new Parser('', input);
  return {
    purgeAllowed: parser.getArray('PURGE_ALLOWED'),
    externalRealIp: parser.getBoolean('EXTERNAL_REAL_IP'),
    realIp: parser.getArray('REAL_IP'),
    limitRate: parser.getString('LIMIT_RATE'),
    limitBurst: parser.getString('LIMIT_BURST'),
    maxCacheSize: parser.getString('MAX_CACHE_SIZE') || '10g',
    dhparamPath:
      parser.getString('DHPARAM_PATH') || '/etc/nginx/generated/dhparam.pem',
    ticketKeyPath:
      parser.getString('TICKET_KEY_PATH') || '/etc/nginx/generated/ticket.key',
    certsPath: parser.getString('CERTS_PATH') || '/etc/nginx/certs',
    sites: await Promise.all(
      getSiteNames().map((domain) => getSiteData(domain, input)),
    ),
    httpExtra: parser.getArray('HTTP_EXTRA'),
    nginxExtra: parser.getArray('NGINX_EXTRA'),
    cors: parser.getBoolean('CORS'),
  };
}
